;; Generated package description from bbdb.el  -*- no-byte-compile: t -*-
(define-package "bbdb" "3.2.2.2.0.20220415.231246" "Big Brother DataBase" '((emacs "24") (cl-lib "0.5")) :commit "715f35b5f53d6bcdcb3754b4f98933df01b57c15" :url "https://elpa.gnu.org/packages/bbdb.html" :maintainer '("Roland Winkler" . "winkler@gnu.org"))
